#File names
JSON_ID = "SessionData.json"
IMG_EXTENSION = [".jpg", ".png", ".JPG", ".PNG", ".JPEG"]
MESH_EXTENSION = [".obj",".fbx" ]
PCD_EXTENSION = [".pcd", ".pts", ".ply"]
TIME_FORMAT = "%Y-%m-%d %H-%M-%S"
METHODS = ["leastDistance","incremental","raycasting", "fpfh", "super4pcs"]